# -*- coding: utf-8 -*-
# Author: jlopes@usp.br
import os 
import numpy as np
from kitconc import wordle


class HeatMap():
    
    def __init__(self,**kwargs):
        self.__path = os.path.dirname(os.path.abspath(__file__))
        self.title = kwargs.get('title','Heat Map')
        self.annotation = kwargs.get('annotation',True)
        self.annotation_color = kwargs.get('annotation_color','k')
        self.fontsize = kwargs.get('fontsize',9)
        self.xrotation = kwargs.get('xrotation',45)
        self.cmap = kwargs.get('cmap','YlOrBr')
        
    
    def plot(self, x_labels,y_labels,values):
        """Shows a heat map plot from lists of values.
        :param  x_labels: A list of string labels for the x axis.
        :type   x_labels: list
        :param  y_labels: A list of string labels for the y axis.
        :type   y_labels: list
        :param  values: A list of int or float values.
        :type   values: list
        """
        import matplotlib
        values = np.array(values)
        fig, ax= matplotlib.pyplot.subplots()
        ax.set_title(self.title)
        ax.set_xticks(np.arange(len(x_labels)))
        ax.set_yticks(np.arange(len(y_labels)))
        ax.set_xticklabels(x_labels,fontsize=self.fontsize)
        ax.set_yticklabels(y_labels,fontsize=self.fontsize)
        matplotlib.pyplot.setp(ax.get_xticklabels(), rotation=self.xrotation, ha="right", rotation_mode="anchor")
        if self.annotation == True:
            for i in range(len(y_labels)):
                for j in range(len(x_labels)):
                    t = ax.text(j, i, values[i, j], ha="center", va="center", color=self.annotation_color)
        matplotlib.pyplot.imshow(values,norm=matplotlib.colors.LogNorm(),interpolation='nearest',cmap=self.cmap)
        matplotlib.pyplot.show()
            
        

class WordCloud(object):
    
    def __init__(self,**kwargs):
        self.__path = os.path.dirname(os.path.abspath(__file__))
        self.limit = kwargs.get('limit',400)
        self.theme = kwargs.get('theme',wordle.THEME_GRAYISH_COLORS)
        self.vertical = kwargs.get('vertical',True)
        self.stoplist = kwargs.get('stoplist',[])
        self.format = kwargs.get('format','plot')
    
    def plot_wordlist(self,wordlist):
        tokens,freq = [],[]
        i = 0
        for row in wordlist.df.itertuples(index=False):
            if row[1] not in self.stoplist:
                tokens.append(row[1])
                freq.append(row[2])
                i+=1
                if i >= self.limit:
                    break 
        vert = 0 
        if self.vertical == True:
            vert = 0.2
        if self.format == 'plot':
            wordle.plot(tokens,freq,theme=self.theme,vertical=vert)
        else:
            wordle.show_image(tokens,freq,theme=self.theme,vertical=vert)
            
    
    def plot_keywords(self,keywords):
        tokens,freq = [],[]
        i = 0
        for row in keywords.df.itertuples(index=False):
            if row[1] not in self.stoplist:
                tokens.append(row[1])
                freq.append(row[3])
                i+=1
                if i >= self.limit:
                    break 
        vert = 0 
        if self.vertical == True:
            vert = 0.2
        if self.format == 'plot':
            wordle.plot(tokens,freq,theme=self.theme,vertical=vert)
        else:
            wordle.show_image(tokens,freq,theme=self.theme,vertical=vert) 
    
    def plot_wfreqinfiles(self,wfreqinfiles):
        tokens,freq = [],[]
        i = 0
        for row in wfreqinfiles.df.itertuples(index=False):
            if row[1] not in self.stoplist:
                tokens.append(row[1])
                freq.append(row[2])
                i+=1
                if i >= self.limit:
                    break 
        vert = 0 
        if self.vertical == True:
            vert = 0.2
        if self.format == 'plot':
            wordle.plot(tokens,freq,theme=self.theme,vertical=vert)
        else:
            wordle.show_image(tokens,freq,theme=self.theme,vertical=vert) 
        

class CollDist(object):
    
    def __init__(self):
        self.__path = os.path.dirname(os.path.abspath(__file__))
    
    def plot_colldist(self,left_freq,right_freq,**kwargs):
        """
        Returns a plot to show the distribution of a collocate.
        Ex.:
        plt = distribution_of_collocate([10,15,30],[20,5,35]) 
        args:
        - left_freq: list of left frequencies
        - right_freq: list of right frequencies 
        kwargs:
        title, label,xlabel, ylabel, left_labels, right_labels
        bar_line_color,bar_fill_color, opacity
        show_values, values_color 
        """
        import matplotlib.pyplot
        # kwargs - colors
        opacity = kwargs.get('opacity',1)
        bar_line_color = kwargs.get('bar_line_color','black')
        bar_fill_color = kwargs.get('bar_fill_color','white')
        # kwargs - labels
        title = kwargs.get('title','Distribution')
        label = kwargs.get('label','...')
        xlabel = kwargs.get('xlabel','Horizon')
        ylabel = kwargs.get('ylabel','Frequency')
        left_labels = kwargs.get('left_labels',['L1','L2','L3','L4','L5'])
        right_labels = kwargs.get('right_labels',['R1','R2','R3','R4','R5'])
        # kwargs - values
        show_values = kwargs.get('show_values',False)
        values_color = kwargs.get('values_color','black')
        # arrange data
        total_left = 5
        total_right = 5
        n_groups = total_left + total_right + 1
        hlabels = []
        hfreqs = []
        i=0
        for freq in left_freq:
            hfreqs.append(freq)
            hlabels.append(left_labels[i])
            i+=1
        if len(left_freq) < 5:
            while len(hfreqs) < 5:
                hfreqs.append(0)
                hlabels.append('-')
        hfreqs = list(reversed(hfreqs))
        hlabels = list(reversed(hlabels))
        hfreqs.append(0)
        hlabels.append('N')
        i=0
        for freq in right_freq:
            hfreqs.append(freq)
            hlabels.append(right_labels[i])
            i+=1
        if len(right_freq) <5:
            while len(hfreqs) < 11:
                hfreqs.append(0)
                hlabels.append('-')
        # make bars
        std_hor = (0,0,0,0,0,0,0,0,0,0,0)
        fig, ax = matplotlib.pyplot.subplots()
        index = np.arange(n_groups)
        bar_width = 0.55
        error_config = {'ecolor': '0.4'}
        ax.bar(index, hfreqs, bar_width,
                        alpha=opacity, edgecolor=bar_line_color,color=bar_fill_color,
                        yerr=std_hor,error_kw=error_config,label=label)
        # format
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title(title)
        ax.set_xticks(index + 0.01)
        ax.set_xticklabels(hlabels)
        ax.legend()
        # check show values
        if show_values == True:
            j = 0
            for i, v in enumerate(hfreqs):
                j+=1
                if v !=0:
                    if j == 6:
                        ax.text((i + (bar_width/2) ) - bar_width, v + 0.3, ' ', color=values_color)
                    else:
                        ax.text((i + (bar_width/2) ) - bar_width, v + 0.3, str(v), color=values_color)
        # show plot
        fig.tight_layout()
        matplotlib.pyplot.show()
        return matplotlib.pyplot
    

class CollGraph(object):
    
    def __init__(self,**kwargs):
        self.title = kwargs.get('title','Collocations')
        self.xlabel = kwargs.get('xlabel','position of the collocate')
        self.ylabel = kwargs.get('ylabel','strength of association')
        self.node = kwargs.get('node','node')
        self.cutoff = kwargs.get('cutoff',0.5)
        self.limit = kwargs.get('limit',20)
        self.stoplist = kwargs.get('stoplist',[])
        self.__load_colors()
        
    
    def __load_colors(self):
        self.__colors = {
            0 : '#000000', 5:  '#0d0d0d', 10: '#1a1a1a', 15: '#262626',
            20: '#333333', 25: '#404040', 30: '#4d4d4d', 35: '#595959',
            40: '#666666', 45: '#737373', 50: '#808080', 55: '#8c8c8c',
            60: '#999999', 65: '#a6a6a6', 70: '#b3b3b3', 75: '#bfbfbf',
            80: '#cccccc', 85: '#d9d9d9', 90: '#e6e6e6', 95: '#f2f2f2', 
            100: '#ffffff', 
            }
        

    def __color(self,p):
        p = 100-p
        arr = np.array([c for c in self.__colors])
        idx = (np.abs(arr-p)).argmin()
        return self.__colors[arr[idx]] 
    
    def __normalize(self,value, max_value,min_value):
        norm = ((value-min_value) / (max_value-min_value)) * 100
        return norm
    
    def __distance(self,norm_stat):
        d = 50 - ( (norm_stat * 50) / 100)
        return d  
    
    def plot_graphcoll(self,collocates):
        data = []
        max_freq = collocates.df['FREQUENCY'].max()
        min_freq = collocates.df['FREQUENCY'].min()
        max_stat = collocates.df['ASSOCIATION'].max()  
        min_stat = collocates.df['ASSOCIATION'].min()
   
        for row in collocates.df.itertuples(index=False):
            if row[5] >= self.cutoff:
                if row[1] not in self.stoplist:
                    if len(row[1]) == 1 and str(row[1]).isalpha()== False:
                        pass 
                    else:
                        word = row[1]
                        norm_freq = self.__normalize(row[2], max_freq, min_freq)
                        norm_stat = self.__normalize(row[5], max_stat, min_stat)
                        distance = self.__distance(norm_stat)
                        if row[3] > row[4]:
                            side = 1
                        elif row[3] < row[4]:
                            side = 2
                        else:
                            side = 0
                        data.append((word,norm_freq,norm_stat,distance,side))
            
            if len(data) >= self.limit:
                break 
        self.plot(data)
                    
    
    def plot(self,data):
        import matplotlib.pyplot
        # set plot title
        matplotlib.pyplot.title(self.title)
        # set limits
        ax = matplotlib.pyplot.gca()
        ax.set_ylim([0,100])
        ax.set_xlim([0,100])
        # define x,y labels
        matplotlib.pyplot.xlabel(self.xlabel)
        matplotlib.pyplot.ylabel(self.ylabel)
        # place node line at the center
        matplotlib.pyplot.plot([50,50],[1,95],color='#9999ff', marker=',',linewidth=1)
        matplotlib.pyplot.plot([50], [95],color='#9999ff',marker=',', markersize=10, label=self.node)
        # set font
        font = {'family': 'Courier New',
        'color':  'black',
        'weight': 'normal',
        'size': 10,
        }
        # set position
        offset = 5
        xr = 100
        xl = 100
        y=100
        i = 0
        for d in data:
            i+=1
            # draw left 
            if d[4] == 1:
                y-= offset
                xl = 49 - d[3]
                matplotlib.pyplot.plot([xl,50],[y,y],color='#999999', marker='.',linewidth=0.1)
                matplotlib.pyplot.plot(xl, y,color=self.__color(d[1]),marker='o', markersize=10)
                matplotlib.pyplot.text((xl-3) - (len(d[0])*1.5),y-1,d[0],fontdict=font)
            # draw right    
            if d[4] == 2:
                y-= offset
                xr = 51 + d[3]
                matplotlib.pyplot.plot([50,xr],[y,y],color='#999999', marker='.',linewidth=0.1)
                matplotlib.pyplot.plot(xr, y,color=self.__color(d[1]),marker='o', markersize=10)
                matplotlib.pyplot.text(xr+2,y-1,d[0],fontdict=font)
            
        # show legend
        matplotlib.pyplot.legend(loc='best')
        matplotlib.pyplot.gca().axes.get_xaxis().set_visible(False)
        matplotlib.pyplot.gca().axes.get_yaxis().set_visible(False)
        # hide frame
        matplotlib.pyplot.box(False)
        # show plot
        matplotlib.pyplot.show()
        



# -*- coding: utf-8 -*-
# Author: jlopes@usp.br
import os
import pickle  
from string import punctuation
from kitconc.core import Config
import collections 
try:
    import nltk 
except:
    nltk = None 


class Models(object):
    
    def __init__(self):
        self.__path = os.path.dirname(os.path.abspath(__file__))
    
    def save_model (self,model,language):
        lm_path = self.__path + '/data/' + language + '_model.pickle'
        if os.path.exists(lm_path):
            os.remove(lm_path)
        with open(lm_path,'wb') as fh:
            pickle.dump(model,fh)
    
    def remove_model(self,language):
        lm_path = self.__path + '/data/' + language + '_model.pickle'
        if os.path.exists(lm_path):
            os.remove(lm_path)
    
    def list_models(self):
        I = 11
        E = 7
        files = os.listdir(self.__path + '/data/')
        language_models =[]
        d = dict()
        for filename in files:
            if filename.endswith('_model.pickle'):
                try:
                    language = filename[:-13]
                    if language not in d:
                        d[language]=None
                except:
                    pass 
            elif filename.startswith('pos_tagger_'):
                try:
                    size = len(filename)
                    m = size - (I+E)
                    language = filename[I:(I+m)]
                    if language not in d:
                        d[language]=None
                except:
                    pass
        for k in d:
            if k != 'language':
                language_models.append(k)
        return language_models 
        
    
    def save_reflist(self,freqlist,language):
        reflist_path = self.__path + '/data/reflist_' + language + '.tab'
        if os.path.exists(reflist_path):
            os.remove(reflist_path)
        s=[]
        for w,f in freqlist:
            s.append('%s\t%s' % (w,f))
        with open(reflist_path,'w',encoding='utf-8') as fh:
            fh.write('\n'.join(s))
    
    def save_stoplist(self,stoplist,language):
        stoplist_path = self.__path + '/data/stolist_' + language + '.tab'
        if os.path.exists(stoplist_path):
            os.remove(stoplist_path)
        with open(stoplist_path,'w',encoding='utf-8') as fh:
            fh.write('\n'.join(stoplist))
    
    def remove_reflist(self,language):
        reflist_path = self.__path + '/data/reflist_' + language + '.tab'
        if os.path.exists(reflist_path):
            os.remove(reflist_path)
        
    def remove_stoplist(self,language):
        stoplist_path = self.__path + '/data/stoplist_' + language + '.tab'
        if os.path.exists(stoplist_path):
            os.remove(stoplist_path)
        
    def dummy_model(self):
        return DummyModel
    
    def spacy_model(self,model_name,disable=[]):
        return spacyModels().load(model_name,disable=disable)
    
    def nltk_create_model(self,source,language,**kwargs):
        sent_tokenizer = kwargs.get('sent_tokenizer',None)
        word_tokenizer = kwargs.get('word_tokenizer',None)
        pos_tagger = kwargs.get('pos_tagger',None) 
        regexp_word = kwargs.get('regexp_word','\w+')
        reflist_path =kwargs.get('reflist',None)
        stoplist_path = kwargs.get('stoplist',None)
        show_progress = kwargs.get('show_progress',False)
        nltk_model =  nltkTrainModel(sent_tokenizer=sent_tokenizer,word_tokenizer=word_tokenizer,pos_tagger=pos_tagger,regexp_word=regexp_word,
                                     reflist=reflist_path,stoplist=stoplist_path, show_progress=show_progress)
        nltk_model.train(source)
        nltk_model.save(language)
        
        

class Token(object):

    def __init__(self,**kwargs):
        self.text = kwargs.get('text',None)
        self.pos_ = kwargs.get('pos',None)
            
class DummyModel(object):
    
    def __init__(self,text):
        self.sents = []
        self.__get_sents(text)
          
    def __get_sents(self,text):
        for c in ['?','!','.']:
            text = text.replace(c, '%s\n' % c)
        for sent in text.split('\n'):
            if len(sent.strip())!=0:
                for p in punctuation:
                    sent = sent.replace(p,' %s ' % p)
                new_sent = []
                for item in sent.split():
                    if len(item.strip())!=0:
                        token = Token(text=item,pos='*')
                        new_sent.append(token)
                self.sents.append(new_sent)
            
    def __iter__(self):
        for sent in self.sents:
            yield sent

                        
class spacyModels(object):
    
    def __init__(self,**kwargs):
        self.disable = kwargs.get('disable',['ner'])
        
    def load(self,model_name,**kwargs):
        self.disable = kwargs.get('disable',['ner'])
        try:
            import spacy 
            return spacy.load(model_name,disable=self.disable)
        except:
            return None 

class nltkTrainModel(object):
    
    def __init__(self,**kwargs):
        self.sent_tokenizer = kwargs.get('sent_tokenizer',None)
        self.word_tokenizer = kwargs.get('word_tokenizer',None)
        self.pos_tagger = kwargs.get('pos_tagger',None) 
        self.regexp_word = kwargs.get('regexp_word','\w+')
        self.reflist_path =kwargs.get('reflist',None)
        self.reflist = None
        self.stoplist = None
        self.stoplist_path = kwargs.get('stoplist',None)
        self.show_progress = kwargs.get('show_progress',False)
    
    def train(self,source_folder):
        # normalize path 
        if not source_folder.endswith('/'):
            source_folder+='/'
        self.source_folder = source_folder
        # reflist 
        self.reflist = []
        if self.reflist_path != None:
            # load reflist from a file
            if os.path.exists(self.reflist_path) and os.path.isfile(self.reflist_path):
                with open(self.reflist_path,'r',encoding='utf-8') as fh:
                    for line in fh:
                        if len(line.strip())!=0:
                            f = line.strip().split('\t')
                            if len(f)>= 2:
                                self.reflist.append((f[0],int(f[1])))
        else:
            # create a reflist from the source corpus
            punct = [c for c in punctuation]
            counter = collections.Counter()
            for word in self.__read_words():
                counter[word.lower()]+=1
            for k,v in counter.most_common():
                if v > 1:
                    if k not in punct:
                        self.reflist.append((k,v))
        # stoplist
        self.stoplist = []
        if self.stoplist_path != None:
            # load stoplist from a file
            if os.path.exists(self.stoplist_path) and os.path.isfile(self.stoplist_path):
                with open(self.stoplist_path,'r',encoding='utf-8') as fh:
                    for line in fh:
                        if len(line.strip())!=0:
                            self.stoplist.append(line.strip())
        else:
            # just use punctuation
            self.stoplist =  [c for c in punctuation]
        # train sent_tokenizer
        if self.show_progress == True:
            print('Training sentence tokenizer...')
        if self.sent_tokenizer == None:
            trainer = nltk.punkt.PunktTrainer()
            trainer.INCLUDE_ALL_COLLOCS = True
            for text in self.__read_texts():
                trainer.train(text,finalize=False)
            self.sent_tokenizer = nltk.punkt.PunktSentenceTokenizer(trainer.get_params())
            trainer = None
        # train word tokenizer
        if self.show_progress == True:
            print('Training word tokenizer...')
        if self.word_tokenizer == None:
            self.word_tokenizer = nltk.RegexpTokenizer(self.regexp_word)
        # train tagger 
        if self.show_progress == True:
            print('Training pos tagger...')
        # get the most common tag 
        counter = collections.Counter()
        for tag in self.__read_tags():
            counter[tag]+=1
        default_tag = counter.most_common(1)[0][0] 
        # get the data for training and testing
        tagged_sents = list(self.__read_tagged_sents())
        cutoff = int(.75 * len(tagged_sents))
        training_sents = tagged_sents[:cutoff]
        test_sents = tagged_sents[cutoff:] 
        tagged_sents = None
        # train sequential taggers   
        default_tagger = nltk.DefaultTagger(default_tag)
        unigram_tagger = nltk.UnigramTagger(training_sents,backoff=default_tagger)
        self.pos_tagger = nltk.BigramTagger(training_sents,backoff=unigram_tagger)
        # print results
        if self.show_progress == True:
            print('Testing tagger...')
        if self.show_progress == True:
            print ('POS tagger precision: %s' % self.pos_tagger.evaluate(test_sents))
        
        
    def save(self,language):
        if self.sent_tokenizer != None and self.word_tokenizer != None and self.pos_tagger != None:
            if self.show_progress == True:
                print('Saving model...')
            config = Config()
            config.add_sent_tokenizer(self.sent_tokenizer, language)
            config.add_tokenizer(self.word_tokenizer, language)
            config.add_tagger(self.pos_tagger, language)
            s = []
            for item in self.reflist:
                s.append('%s\t%s' % (item[0],item[1]))
            config.add_reflist('\n'.join(s), language)
            s = None
            config.add_stoplist('\n'.join(self.stoplist), language)    
            if self.show_progress == True:
                print('OK.')
        else:
            if self.show_progress == True:
                print('Kitconc cannot save the model. Trained components are missing.')
         
    
    
    def __read_texts(self):
        files = os.listdir(self.source_folder)
        for filename in files:
            text = []
            with open(self.source_folder + filename,'r', encoding='utf-8') as fh:
                for line in fh:
                    if len(line.strip())!=0:
                        raw_sent = [ self.__str2tuple(token)[0] for token in line.strip().split()]
                        text.append(self.__norm_punct(' '.join(raw_sent)))
            yield '\n'.join(text)
        
    def __read_tagged_sents(self):
        files = os.listdir(self.source_folder)
        for filename in files:
            with open(self.source_folder + filename,'r', encoding='utf-8') as fh:
                for line in fh:
                    if len(line.strip())!=0:
                        tagged_sent = [ self.__str2tuple(token) for token in line.strip().split()]
                        yield tagged_sent
    
    
    def __read_raw_sents(self):
        files = os.listdir(self.source_folder)
        for filename in files:
            with open(self.source_folder + filename,'r', encoding='utf-8') as fh:
                for line in fh:
                    if len(line.strip())!=0:
                        raw_sent = [ self.__str2tuple(token)[0] for token in line.strip().split()]
                        yield self.__norm_punct(' '.join(raw_sent))
    
    def __read_tags(self):
        files = os.listdir(self.source_folder)
        for filename in files:
            with open(self.source_folder + filename,'r', encoding='utf-8') as fh:
                for line in fh:
                    if len(line.strip())!=0:
                        tag_sent = [ self.__str2tuple(token)[1] for token in line.strip().split()]
                        for tag in tag_sent:
                            yield tag
    
    def __read_words(self):
        files = os.listdir(self.source_folder)
        for filename in files:
            with open(self.source_folder + filename,'r', encoding='utf-8') as fh:
                for line in fh:
                    if len(line.strip())!=0:
                        word_sent = [self.__str2tuple(token)[0] for token in line.strip().split()]
                        for word in word_sent:
                            yield word
    
    def __str2tuple(self,s):
        sep='/'
        loc = s.rfind(sep)
        if loc >= 0:
            return (s[:loc], s[loc + len(sep) :])
        else:
            return (s, 'N')
    
    def __norm_punct(self,s):
        s = s.replace(' .','.')
        s = s.replace(' ?','?')
        s = s.replace(' !','!')
        s = s.replace(' ,',',')
        s = s.replace(' ;',';')
        s = s.replace(' :',':')
        return s 
    
        


            
             




# -*- coding: utf-8 -*-
# Author: jlopes@usp.br
import os 
import pickle
import numpy as np
import re  


def load_dict(dict_path):
    with open(dict_path,'rb') as fh:
        d = pickle.load(fh)
    return d 

def save_dict(dict_d,dict_path):
    with open(dict_path,'wb') as fh:
        pickle.dump(dict_d,fh)

def count_words(npy_path):
    counter = dict()
    files = os.listdir(npy_path)
    for filename in files:
        arr=np.load(npy_path + filename)
        i,c = np.unique(arr[:,0], return_counts=True)
        for x in range(i.shape[0]):
            if i[x] in counter:
                counter[i[x]]+=c[x]
            else:
                counter[i[x]]=c[x]
    return counter
    
def make_wordlist(workspace,corpus_name,language,lowercase=True):
    # check wordlist saved 
    if os.path.exists(workspace + corpus_name + '/data/idx/wordlistlc.pickle') == True and lowercase == True:
        with open(workspace + corpus_name + '/data/idx/wordlistlc.pickle','rb') as fh:
            wordlist = pickle.load(fh)
        info = list()
        with open(workspace + corpus_name + '/data/idx/info.pickle','rb') as fh:
            info = pickle.load(fh)
        return (info[0],info[1],info[2],info[3],wordlist)  
    
    # compile regex 
    punct = re.compile('^\W+$')
    
    # make or load wordlist count
    if (os.path.exists(workspace+corpus_name+'/data/idx/wordlist.pickle') == False):
        word_count = count_words(workspace+corpus_name + '/data/npy/')
        save_dict(word_count,workspace+corpus_name+'/data/idx/wordlist.pickle')
    else:
        word_count = load_dict(workspace+corpus_name+'/data/idx/wordlist.pickle')
    
    # load words dictionary
    dict_words = load_dict(workspace+corpus_name+'/data/idx/words.pickle')
    
    # lowercase
    if lowercase == True:
        new_word_count = dict()
        tokens = 0
        for k, v in sorted(word_count.items(), key=lambda item: item[1],reverse=True):
            tokens+=v
            s = dict_words[k].lower()
            if s in new_word_count:
                new_word_count[s]+=v 
            else:
                new_word_count[s]=v
        word_count = None
        dict_words = None
        types = len(new_word_count)
        total = float(tokens)
        wordlist = list()
        n= 0
        hapax = 0
        for s, v in sorted(new_word_count.items(), key=lambda item: item[1],reverse=True):
            if punct.match(s) == None:
                n+=1
                p = (v/total)*100  
                wordlist.append((n,s,v,p))
                if v == 1:
                    hapax+=1
    else:
        for k, v in sorted(word_count.items(), key=lambda item: item[1],reverse=True):
            tokens+=v
        total = float(tokens)
        types = len(word_count)
        wordlist = list()
        for k, v in sorted(word_count.items(), key=lambda item: item[1],reverse=True):
            s = dict_words[k]
            if punct.match(s) == None:
                n+=1
                p = (v/total)*100  
                wordlist.append((n,s,v,p))
                if v == 1:
                    hapax+=1
        word_count = None
        dict_words = None
    type_token = (types/float(tokens))*100
    
    if os.path.exists(workspace + corpus_name + '/data/idx/wordlistlc.pickle') == False and lowercase == True:
        with open(workspace + corpus_name + '/data/idx/wordlistlc.pickle','wb') as fh:
            pickle.dump(wordlist,fh)
        info = [tokens,types,type_token,hapax]
        with open(workspace + corpus_name + '/data/idx/info.pickle','wb') as fh:
            pickle.dump(info,fh)
        with open(workspace + corpus_name + '/info.tab','a',encoding='utf-8') as fh:
            fh.write('\nTokens:\t%s\nTypes:\t%s\nType/token:\t%s' % (tokens,types,type_token))
    return (tokens,types,type_token,hapax,wordlist)  
    
    



# -*- coding: utf-8 -*-
# Author: jlopes@usp.br
import os,sys  
import math
import time  
import string 

def __progress(count, total, suffix=''):
        bar_len = 30
        filled_len = int(round(bar_len * count / float(total)))
        percents = round(100.0 * count / float(total), 1)
        bar = '=' * filled_len + '-' * (bar_len - filled_len)
        sys.stdout.write('[%s] %s%s ...%s\r' % (bar, percents, '%', suffix))
        try:
            sys.stdout.flush()  
        except:
            pass
        
def keywords_reference(filename):
    def readlines(filename):
        i = 0
        with open(filename,'r',encoding='utf-8') as fh:
            for line in fh:
                if len(line.strip())!=0:
                    if i != 0:
                        fields = line.strip().split('\t')
                        if len(fields) >=4:
                            yield (fields[1],fields[2])
                    i+=1
    tokens = 0
    reflist = {}
    freqlist = readlines(filename)
    for item in freqlist:
        tokens+= int(item[1])
        if item[0] not in reflist:
            reflist[item[0]]=int(item[1])
    return (reflist,tokens) 
        

def load_reference_wordlist(filename):
        reflist = {}
        i = 0
        with open(filename,'r','utf-8') as fh:
            for line in fh:
                if len(line.strip())!=0:
                    if i != 0:
                        fields = line.strip().split('\t')
                        if len(fields) >=4:
                            if fields[1] not in reflist:
                                reflist[fields[1]]=int(fields[2])
                    i+=1
        return reflist

def sent2conll2000(sent):
    s = []
    for token in sent:
        s.append(token[0] + ' ' + token[1] + ' O')
    return '\n'.join(s)   

    
def chi_square(freq_stdc,freq_refc,tk_stdc,tk_refc):
    a = freq_stdc 
    b = freq_refc 
    c = tk_stdc - a 
    d = tk_refc - b 
    N = a + b + c + d
    chi = N * (a*d-b*c)** 2/((a+b)*(c+d)*(a+c)*(b+d))
    chi = round(chi,2) 
    return chi 

def observed_expected(ab,a,b,N, h=1):
    O = ab / float(N)
    E = (a / float(N)) * (b / float(N))
    E =  E * h
    OE = O/E
    return OE 

def mutual_information(ab,a,b,N, h=1):
    O = ab / float(N)
    E = (a / float(N)) * (b / float(N))
    E =  E * h
    OE = O/E
    I = math.log(OE,2)
    I = round(I,2)
    return I 

def tscore(ab,a,b,N, h=1):
    O = ab / float(N)
    E = (a / float(N)) * (b / float(N))
    E =  E * h
    T = (O-E) / (math.sqrt(ab)/float(N))
    T = round(T,2)
    return T

def ll(freq_stdc,freq_refc,tk_stdc,tk_refc):
    """Calculates the log-likelihood value"""
    O = float(freq_stdc+freq_refc)
    N1 = float(tk_stdc)
    N2 = float(tk_refc)
    E1 = N1*O/(N1+N2) 
    E2 = N2*O/(N1+N2)
    if freq_stdc < 1 or E1 < 1:
        v1 = 0
    else:
        v1 = math.log(freq_stdc/E1)
    if freq_refc < 1 or E2 < 1:
        v2 = 0
    else:
        v2 = math.log(freq_refc/E2)
    LL = round(2 * ((freq_stdc * v1)+(freq_refc * v2)),2)
    Norm_stdc = freq_stdc/float(tk_stdc)
    Norm_refc = freq_refc/float(tk_refc)
    if Norm_stdc == 0:
        Norm_stdc = 0.5 / float(tk_stdc)
    if Norm_refc == 0:
        Norm_refc = 0.5 / float(tk_refc)
    if Norm_stdc < Norm_refc:
        LL = - LL
    return LL


def logRatio(freq_stdc,freq_refc,tk_stdc,tk_refc):
    Norm_stdc = freq_stdc/float(tk_stdc)
    Norm_refc = freq_refc/float(tk_refc)
    if Norm_stdc == 0:
        Norm_stdc = 0.5 / float(tk_stdc)
    if Norm_refc == 0:
        Norm_refc = 0.5 / float(tk_refc)
    lr = math.log2((Norm_stdc/Norm_refc))
    return lr

def dump(obj,filename):
    import pickle
    with open(filename,'wb') as fh:
        pickle.dump(obj,fh)

def load(filename):
    import pickle 
    with open(filename,'rb') as fh:
        obj = pickle.load(fh)
    return obj 

def file2utf8(source_file,target_file,source_encoding='mbcs'):
    import codecs
    BLOCKSIZE = 1048576 # or some other, desired size in bytes
    with codecs.open(source_file, "r", source_encoding,errors="surrogateescape") as sourceFile:
        with codecs.open(target_file, "w", "utf-8") as targetFile:
            while True:
                contents = ''.join(filter(lambda x: x in string.printable, sourceFile.read(BLOCKSIZE)))
                if not contents:
                    break
                targetFile.write(contents)
                

def file2utf8b(source_file,target_file,source_encoding='mbcs'):
    import codecs 
    with open(source_file,'rb') as fh1:
        with codecs.open(target_file,'w','utf-8') as fh2:
            while True:
                contents = fh1.read()
                if not contents:
                    break
                fh2.write(contents.decode(source_encoding))

def files2utf8(source_folder,target_folder,source_encoding='mbcs',show_progress=False):
    # time start
    if show_progress == True:
        print('Running...')
        t0 = time.time()
    if source_folder.endswith('/')==False:
        source_folder = source_folder + '/'
    if target_folder.endswith('/') == False:
        target_folder = target_folder + '/'
    if os.path.exists(target_folder)==False:
        os.mkdir(target_folder)
    files = os.listdir(source_folder)
    i = 0
    total = len(files)
    for filename in files:
        i+=1
        if os.path.isfile(source_folder + filename)==True:
            source_encoding = encoding_get(source_folder + filename)
            if source_encoding != 'utf-8':
                file2utf8b(source_folder + filename,target_folder + filename,source_encoding)
            else:
                with open (source_folder + filename,'r', encoding='utf-8') as fh:
                    content = fh.read()
                with open(target_folder + filename,'w', encoding='utf-8') as fh:
                    fh.write(content)
            if show_progress == True:
                __progress(i, total, '')
    # time end
    if show_progress == True:
        t1 = time.time()
        total_time = round(t1 - t0,2) 
        print('')
        print('Total time: %s seconds' % total_time) 

def encoding_get(filename):
    import chardet
    with open (filename,'rb') as fh:
        dados = b''.join([l for l in fh])
        return chardet.detect(dados)['encoding']

    






# Designed and implemented by: Hayk Aleksanyan
# based on the approach by Jonathan Feinberg, see http://static.mrfeinberg.com/bv_ch03.pdf

import os 
from PIL import Image, ImageColor, ImageFont, ImageDraw
import math
import random
from kitconc import wordle_spirals as SP
from kitconc import wordle_bbox
from kitconc import wordle_trees
from kitconc import wordle_colorhandler as CH
import tempfile 


# vertical constant:
VERTICAL_DISABLED = 0

# stretch sizes constants:
STRETCH_SIZE_10 = 0.1
STRETCH_SIZE_25 = 0.25
STRETCH_SIZE_50 = 0.5
STRETCH_SIZE_75 = 0.75
STRETCH_SIZE_95 = 0.9
STRETCH_SIZE_100 = 1

# color themes constants:
THEME_RANDOM_COLORS = 0
THEME_JET_COLORS = 1
THEME_GRAYISH_COLORS = 2
THEME_FIXED_COLORS = 3
THEME_BRFLAG = 4

#from cython_wordle.main import interActive
interActive = False
# constants:
TOKENS_TO_USE = 400       # number of different tokens to use in the wordle
STAY_AWAY = 2             # force any two words to stay at least this number of pixels away from each other
FONT_SIZE_MIN = 10        # the smallest font of a word
FONT_SIZE_MAX = 300       # the largest font of a word, might go slightly above this value
DESIRED_HW_RATIO = 0.618  # height/widht ratio of the canvas
QUADTREE_MINSIZE = 5      # minimal height-width of the box in quadTree partition
FONT_NAME = os.path.dirname(os.path.abspath(__file__)) +  "/data/ubuntu.ttf"   # the font (true type) used to draw the word shapes



class Token:
    """
        encapsulates the main information on a token into a single class
        Token here represents a word to be placed on canvas for the final wordle Image

        most of the attributes are filled with functions during processing of the tokens
    """

    def __init__(self, word, fontSize = 10, drawAngle = 0):
        self.word = word
        self.fontSize = fontSize      # an integer
        self.drawAngle = drawAngle    # an integer representing the rotation angle of the image; 0 - for NO rotation
        self.imgSize = None           # integers (width, height) size of the image of this word with the given fontSize
        self.quadTree = None          # the quadTree of the image of this word with the above characteristics
        self.place = None             # tuple, the coordinate of the upper-left corner of the token on the final canvas
        self.color = None             # the fill color on canvas (R, G, B) triple


def proposeCanvasSize(normalTokens):
    """
      Given a list of normalized tokens we propose a canvase size (width, height)
      based on the areas covered by words. The areas are determined via leaves of the
      corresponding quadTrees.

      It is assumed that tokens come sorted (DESC), i.e. bigger sized words come first.
      We use this assumption when enforcing the canvas size to be larger
      than total area of the bounding boxes of the first few tokens.
    """

    area = 0         # the area covered by the quadTrees, i.e. the actual shape of the token
    boxArea = []     # the areas covered by the bounding box of the tokens

    for token in normalTokens:
        area += token.quadTree.areaCovered()
        boxArea.append( wordle_trees.rectArea(token.quadTree.root.value) )

    ensure_space = 5    # force the sum of the total area to cover at least the first @ensure_space tokens

    total = area + sum ( boxArea[:ensure_space] )
    w = int( math.sqrt(total/DESIRED_HW_RATIO) ) + 1
    h = int(DESIRED_HW_RATIO*w) + 1

    #print('Ratio of the area covered by trees over the total area of bounding boxes of words',  area/sum(boxArea))
    return w, h

def randomFlips(n, p):
    """
     Return an array of length n of random bits {0, 1} where Probability(0) = p and Probability(1) = 1 - p
     this is used for randomly selecting some of the tokens for vertical placement.
    """
    ans = n*[0]
    for i in range(n):
        x = random.random()
        if x > p:
            ans[i] = 1

    return ans


def normalizeWordSize(tokens, freq, N_of_tokens_to_use, horizontalProbability = 1.0):
    """
     (linearly) scale the font sizes of tokens to a new range depending on the ratio of the current min-max
     and take maximum @N_of_tokens_to_use of these tokens
     allow some words to have vertical orientation defined by @horizontalProbability
    """

    words = tokens[:N_of_tokens_to_use]
    sizes = freq[:N_of_tokens_to_use]

    normalTokens = [ ] # the list of Tokens to be returned

    # scale the range of sizes; the scaling rules applied below are fixed from some heuristic considerations
    # the user of this code is welcome to apply their own reasoning

    a, b = min(sizes), max(sizes)
    if a == b:
        sizes = len(sizes)*[30]
    else:
        if b <= 8*a:
            m, M = 15, 1 + int(20*b/a)
        elif b <= 16*a:
            m, M = 14, 1 + int(18*b/a)
        elif b <= 32*a:
            m, M = 13, 1 + int(9*b/a)
        elif b <= 64*a:
            m, M = 11, 1 + int(4.7*b/a)
        else:
            m, M = FONT_SIZE_MIN, FONT_SIZE_MAX

        sizes = [  int(((M - m )/(b - a))*( x - a ) + m )  for x in sizes ]

    # allow some vertical placement; the probability is defined by the user
    flips = randomFlips(len( words ), horizontalProbability )
    for i in range(len(sizes)):
        normalTokens.append( Token( words[i], sizes[i], 0 if flips[i] == 0 else 90 ) )

    return normalTokens



def drawWord(token, useColor = False):
    """
      gets an instance of Token class and draws the word it represents
      returns an image of the given word in the given font size
      the image is NOT cropped
    """

    font = ImageFont.truetype(FONT_NAME, token.fontSize)
    w, h = font.getsize(token.word)

    im = Image.new('RGBA', (w,h), color = None)
    draw = ImageDraw.Draw(im)
    if useColor == False:
        draw.text((0, 0), token.word, font = font)
    else:
        draw.text((0, 0), token.word, font = font, fill = token.color)

    if token.drawAngle != 0:
        im = im.rotate( token.drawAngle,  expand = 1)

    return im


def drawOnCanvas(normalTokens, canvas_size,theme=0):
    """
       given a list of tokens and a canvas size, we put the token images onto the canvas
       the places of each token on this canvas has already been determined during placeWords() call.

       Notice, that it is not required that the @place of each @token is inside the canvas;
       if necessary we may enlarge the canvas size to embrace these missing images
    """

    c_W,c_H = canvas_size        # the suggested canvas size, might change here

    # there can be some positions of words which fell out of the canvas
    # we first need to go through these exceptions (if any) and expand the canvas and (or) shift the coordinate's origin.

    X_min, Y_min = 0, 0

    for i, token in enumerate(normalTokens):
        if token.place == None:
            continue

        if X_min > token.place[0]:  X_min = token.place[0]
        if Y_min > token.place[1]:  Y_min = token.place[1]

    x_shift, y_shift = 0, 0
    if X_min < 0:   x_shift = -X_min
    if Y_min < 0:   y_shift = -Y_min

    X_max , Y_max = 0, 0
    for i, token in enumerate(normalTokens):
        if token.place == None:
            continue

        token.place = ( token.place[0] + x_shift, token.place[1] + y_shift )
        if X_max < token.place[0] + token.imgSize[0]:
            X_max = token.place[0] + token.imgSize[0]
        if Y_max < token.place[1] + token.imgSize[1]:
            Y_max = token.place[1] + token.imgSize[1]

    c_W = max(c_W, X_max)
    c_H = max(c_H, Y_max)

    im_canvas = Image.new('RGBA', (c_W + 10 ,c_H + 10 ), color = None )
    im_canvas_white = Image.new('RGBA', (c_W + 10 ,c_H + 10 ), color = (255,255,255,255) )

    # decide the background color with a coin flip; 0 -for white; 1 - for black (will need brigher colors)
    background = random.randint(0,0) # I've changed to be always white

    dd = ImageDraw.Draw(im_canvas)
    if background == 0: # white
        dd_white = ImageDraw.Draw(im_canvas_white)

    #COLOR THEMES (CUSTOMIZED)
    # add color to each word to be placed on canvas, pass on the background info as well
    if theme == 0:
        background = random.randint(0,1)
        CH.randomColors(normalTokens, background)
    elif theme == 1:
        background = 0
        CH.jetColors(normalTokens)
    elif theme == 2:
        background = 0
        CH.grayishRandomColors (normalTokens)
    elif theme == 3:
        background = background = 0
        CH.chooseFromFixedSchemes(normalTokens)
    elif theme >=4:
        background = background = 0
        CH.chooseFromCustom(normalTokens) # custom themes
        
        

    for i, token in enumerate(normalTokens):
        if token.place == None:
            #print('the word <' + token.word + '> was skipped' )
            continue

        font1 = ImageFont.truetype(FONT_NAME, token.fontSize)
        c = token.color

        if token.drawAngle != 0:
            # place vertically, since PIL does support drawing text in vertical orientation,
            # we first draw the token in a temporary image, the @im, then past that at the location of
            # the token on the canvas; this might introduce some rasterization for smaller fonts
            im = drawWord(token, useColor = True)
            im_canvas.paste(im,  token.place, im )
            if background == 0:
                im_canvas_white.paste(im,  token.place, im )
        else:
            dd.text( token.place, token.word, fill = c,  font = font1 )
            if background == 0:
                dd_white.text( token.place, token.word, fill = c,  font = font1 )


    margin_size = 10 # the border margin size
    box = im_canvas.getbbox()

    if background == 0:
        # white background
        im_canvas_1 = Image.new('RGBA', ( box[2] - box[0] + 2*margin_size, box[3] - box[1] + 2*margin_size ), color = (100,100,100,100)  )
        im_canvas_1.paste( im_canvas_white.crop(box), ( margin_size, margin_size, margin_size + box[2] - box[0], margin_size + box[3] - box[1] ) )
    else:
        # black background
        im_canvas_1 = Image.new('RGB', ( box[2] - box[0] + 2*margin_size, box[3] - box[1] + 2*margin_size ), color = (0,0,0)  )
        im_canvas_1.paste( im_canvas.crop(box), ( margin_size, margin_size, margin_size + box[2] - box[0], margin_size + box[3] - box[1] ) )

    return im_canvas_1

def createQuadTrees(normalTokens):
    """
        given a list of tokens we fill their quadTree attributes and cropped image size
    """
    for i, token in enumerate(normalTokens):
        im_tmp = drawWord(token)
        T = wordle_bbox.getQuadTree( im_tmp , QUADTREE_MINSIZE, QUADTREE_MINSIZE )
        T.compress()
        im_tmp = im_tmp.crop(im_tmp.getbbox())
        token.quadTree = T
        token.imgSize = im_tmp.size

def placeWords(normalTokens):
    """
      gets a list of tokens and their frequencies
      executes the placing strategy and
      returns canvas size, locations of upper-left corner of words and words' sizes
    """

    # 1. we first create the QuadTrees for all words and determine a size for the canvas
    word_img_path = [] # shows the path passed through the spiral before hitting a free space
    # create the quadTrees and collect sizes (width, height) of the cropped images of the words
    createQuadTrees(normalTokens)
    #2. We now find places for the words on our canvas
    c_W, c_H = 4000, 4000
    #c_W, c_H = 2000, 1000
    #c_W, c_H = 3000, 1500
    #c_W, c_H = 1000, 1000
    #3a. we start with the 1st word
    ups_and_downs = [ random.randint(0,20)%2  for i in range( len(normalTokens) )]
    for i, token in enumerate(normalTokens):
        a = 0.2                # the parameter of the spiral
        if ups_and_downs[i] == 1:
            # add some randomness to the placing strategy
            a = -a
        # determine a starting position on the canvas of this token, near half of the width of canvas
        w, h =   random.randint( int(0.3*c_W), int(0.7*c_W) ) ,  (c_H >> 1) - (token.imgSize[1] >> 1)
        if w < 0 or w >= c_W:
            w = c_W >> 1
        if h < 0 or h >= c_H:
            h = c_H >> 1
        if ups_and_downs[i] == 0:
            A = SP.Archimedian(a).generator
        else:
            A = SP.Rectangular(2, ups_and_downs[i]).generator
        dx0, dy0 = 0, 0
        place1 = (w, h)
        word_img_path.append( (w,h) )
        last_hit_index = 0 # we cache the index of last hit
        iter_ = 0
        start_countdown = False
        max_iter = 0
        for dx, dy in A:
            w, h = place1[0] + dx, place1[1] + dy
            if start_countdown == True:
                max_iter -= 1
                if max_iter == 0:
                    break
            else:
                iter_ += 1
            if ( w < 0 or w >= c_W or h < 0 or h > c_H ):
                #  the shape has fallen outside the canvas
                if start_countdown == False:
                    start_countdown = True
                    max_iter  = 1 + 10*iter_
            place1 = ( w, h )
            collision = False
            if last_hit_index < i:
                j = last_hit_index
                if normalTokens[j].place != None:
                    collision = wordle_bbox.collisionTest( token.quadTree, normalTokens[j].quadTree, place1, normalTokens[j].place, STAY_AWAY)
            if collision == False:
                # NO collision with the cached index
                for j in range( i ): # check for collisions with the rest of the tokens
                    if ((j != last_hit_index) and (normalTokens[j].place != None)):
                        if wordle_bbox.collisionTest(token.quadTree, normalTokens[j].quadTree, place1, normalTokens[j].place, STAY_AWAY) == True:
                            collision = True
                            last_hit_index = j
                            break # no need to check with the rest of the tokens, try a new position now
            if collision == False:
                if wordle_bbox.insideCanvas( token.quadTree , place1, (c_W, c_H) ) == True:
                    # at this point we have found a place inside the canvas where the current token has NO collision
                    # with the already placed tokens; The search has been completed.
                    token.place = place1
                    break   # breaks the spiral movement
                else:
                    if token.place == None:
                        # even though this place is outside the canvas, it is collision free and we
                        # store it in any case to ensure that the token will be placed
                        token.place = place1
    return c_W, c_H


def plot(tokens,freq,**kwargs):
    """
    Shows wordle in a matplotlib figure.
    themes: 0 random colors; 1 jet (heatmap); 2 grayish; 3 fixed colors.
    """
    import matplotlib.pyplot as plt
    import matplotlib.image as mpimg
    # args
    stretch = kwargs.get('strech',1.0)
    theme = kwargs.get('theme',2)
    vpos = kwargs.get('vertical',.1)
    hpos = kwargs.get('horizontal',1.0)
    verticalProbability = vpos    # the probability of placing some words vertically
    horizontalProbability = hpos - verticalProbability # the probability of placing some words horizontally
    normalTokens =  normalizeWordSize(tokens, freq, TOKENS_TO_USE, horizontalProbability)
    canvas_W, canvas_H = placeWords(normalTokens)
    wordle = drawOnCanvas(normalTokens, (canvas_W, canvas_H ),theme)
    # resize
    if stretch < 1:
        p = stretch * 100
        width = int( (wordle.width * p)/100)
        height = int((wordle.height * p)/100)
        wordle = wordle.resize((width,height), Image.BICUBIC) 
    temp = tempfile.TemporaryFile()
    try:
        wordle.save(temp.name + '.png')
        img=mpimg.imread(temp.name + '.png')
        imgplot = plt.imshow(img,interpolation='bessel')
        plt.axis('off')
        plt.show()
        os.remove(temp.name + '.png')
    finally:
        temp.close()

def show_image(tokens,freq,**kwargs):
    """
    Show wordle image in the default OS image viewer.
    """
    # args
    stretch = kwargs.get('strech',1.0)
    theme = kwargs.get('theme',2)
    vpos = kwargs.get('vertical',.1)
    hpos = kwargs.get('horizontal',1.0)
    verticalProbability = vpos    # the probability of placing some words vertically
    horizontalProbability = hpos - verticalProbability # the probability of placing some words horizontally
    normalTokens =  normalizeWordSize(tokens, freq, TOKENS_TO_USE, horizontalProbability)
    canvas_W, canvas_H = placeWords(normalTokens)
    wordle = drawOnCanvas(normalTokens, (canvas_W, canvas_H ),theme )
    # resize
    if stretch < 1:
        p = stretch * 100
        width = int((wordle.width * p)/100)
        height = int((wordle.height * p)/100)
        wordle = wordle.resize((width,height), Image.BICUBIC)
    wordle.show('Word cloud')
    

def save_image(tokens,freq,filename,**kwargs):
    """
    Show wordle image in the default OS image viewer.
    """
    # args
    stretch = kwargs.get('strech',1.0)
    theme = kwargs.get('theme',2)
    vpos = kwargs.get('vertical',.1)
    hpos = kwargs.get('horizontal',1.0)
    verticalProbability = vpos    # the probability of placing some words vertically
    horizontalProbability = hpos - verticalProbability # the probability of placing some words horizontally
    normalTokens =  normalizeWordSize(tokens, freq, TOKENS_TO_USE, horizontalProbability)
    canvas_W, canvas_H = placeWords(normalTokens)
    wordle = drawOnCanvas(normalTokens, (canvas_W, canvas_H ),theme)
    # resize
    if stretch < 1:
        p = stretch * 100
        width = int( (wordle.width * p)/100)
        height = int((wordle.height * p)/100)
        wordle = wordle.resize((width,height), Image.BICUBIC)
    wordle.save(filename)

    
    
    




import os 
import numpy as np
cimport numpy as np
cimport cython
DTYPE = np.int
ctypedef np.int_t DTYPE_t 
import pickle 
import re 
import collections 
import tempfile 
import shutil

cdef dict load_dict(str dict_path):
    cdef:
        dict d  
    with open(dict_path,'rb') as fh:
        d = dict(pickle.load(fh))
    return d

cdef object parse_search(str workspace,str corpus_name,object pos):
    """Parses the input search and returns a tuple with encoded word and pos node.
    """
    cdef:
        dict dict_tags
        list p 
        str t 
        object encoded_pos 
    # load dictionaries for encoding words to numbers
    dict_tags = load_dict(workspace+corpus_name+'/data/idx/tags.pickle')
    dict_tags = {v: k for k, v in dict_tags.items()}
    # parse pos 
    if pos != None:
        p = []
        if type(pos) == list:
            for t in pos:
                if len(t.strip())!= 0:
                    p.append(t.strip())
        else:
            for t in pos.strip().split(' '):
                if len(t.strip())!=0:
                    p.append(t.strip())
    # encode pos
    if pos != None:
        encoded_pos = []
        for t in p:
            if t in dict_tags:
                arr = np.array(dict_tags[t],dtype=int)
                encoded_pos.append(arr)
            else: # if any tag has no match, return None
                encoded_pos = None
                break 
    else:
        encoded_pos = None
    return encoded_pos


cdef object unigrams(np.ndarray arr,object encoded_pos):
    cdef:
        object dq
        object counter 
        int i  
    if encoded_pos != None:
        dq = collections.deque(maxlen=1)
        counter = collections.Counter()
        for i in range(0,arr.shape[0]):
            if int(arr[i][1]) in encoded_pos:
                dq.append(int(arr[i][0]))
                counter[dq[0]]+=1 
    else:
        dq = collections.deque(maxlen=1)
        counter = collections.Counter()
        for i in range(0,arr.shape[0]):
            dq.append(int(arr[i]))
            counter[dq[0]]+=1
    return counter
    
cdef object bigrams(np.ndarray arr,object encoded_pos):
    cdef:
        object dq
        object dqt 
        int i 
        object counter  
    if encoded_pos != None:
        dq = collections.deque(maxlen=2)
        dqt = collections.deque(maxlen=2)
        dq.append(int(arr[0][0]))
        dqt.append(int(arr[0][1]))
        counter = collections.Counter()
        for i in range(1,arr.shape[0]):
            dq.append(int(arr[i][0]))
            dqt.append(int(arr[i][1]))
            if dqt[0] in encoded_pos[0] and dqt[1] in encoded_pos[1]:
                counter[(dq[0],dq[1])]+=1
    else:
        dq = collections.deque(maxlen=2)
        dq.append(int(arr[0]))
        counter = collections.Counter()
        for i in range(1,arr.shape[0]):
            dq.append(int(arr[i]))
            counter[(dq[0],dq[1])]+=1
    return counter 

cdef object trigrams(np.ndarray arr,object encoded_pos):
    cdef:
        object dq
        object dqt 
        int i 
        object counter
    if encoded_pos != None:
        dq = collections.deque(maxlen=3)
        dqt = collections.deque(maxlen=3)
        dq.append(int(arr[0][0]))
        dq.append(int(arr[1][0]))
        dqt.append(int(arr[0][1]))
        dqt.append(int(arr[1][1]))
        counter = collections.Counter()
        for i in range(2,arr.shape[0]):
            dq.append(int(arr[i][0]))
            dqt.append(int(arr[i][1]))
            if dqt[0] in encoded_pos[0] and dqt[1] in encoded_pos[1] and dqt[2] in encoded_pos[2]:
                counter[(dq[0],dq[1],dq[2])]+=1
    else:
        dq = collections.deque(maxlen=3)
        dq.append(int(arr[0]))
        dq.append(int(arr[1]))
        counter = collections.Counter()
        for i in range(2,arr.shape[0]):
            dq.append(int(arr[i]))
            counter[(dq[0],dq[1],dq[2])]+=1
        
    return counter

cdef object quadrigrams(np.ndarray arr,object encoded_pos):
    cdef:
        object dq
        object dqt 
        int i 
        object counter
    if encoded_pos != None:
        dq = collections.deque(maxlen=4)
        dqt = collections.deque(maxlen=4)
        dq.append(int(arr[0][0]))
        dq.append(int(arr[1][0]))
        dq.append(int(arr[2][0]))
        dqt.append(int(arr[0][1]))
        dqt.append(int(arr[1][1]))
        dqt.append(int(arr[2][1]))
        counter = collections.Counter()
        for i in range(3,arr.shape[0]):
            dq.append(int(arr[i][0]))
            dqt.append(int(arr[i][1]))
            if dqt[0] in encoded_pos[0] and dqt[1] in encoded_pos[1] and dqt[2] in encoded_pos[2] and dqt[3] in encoded_pos[3]:
                counter[(dq[0],dq[1],dq[2],dq[3])]+=1
    else:
        dq = collections.deque(maxlen=4)
        dq.append(int(arr[0]))
        dq.append(int(arr[1]))
        dq.append(int(arr[2]))
        counter = collections.Counter()
        for i in range(3,arr.shape[0]):
            dq.append(int(arr[i]))
            counter[(dq[0],dq[1],dq[2],dq[3])]+=1
        
    return counter 
    
cdef object quinquegrams(np.ndarray arr,object encoded_pos):
    cdef:
        object dq
        object dqt 
        int i 
        object counter
    if encoded_pos != None:
        dq = collections.deque(maxlen=5)
        dqt = collections.deque(maxlen=5)
        dq.append(int(arr[0][0]))
        dq.append(int(arr[1][0]))
        dq.append(int(arr[2][0]))
        dq.append(int(arr[3][0]))
        dqt.append(int(arr[0][1]))
        dqt.append(int(arr[1][1]))
        dqt.append(int(arr[2][1]))
        dqt.append(int(arr[3][1]))
        counter = collections.Counter()
        for i in range(4,arr.shape[0]):
            dq.append(int(arr[i][0]))
            dqt.append(int(arr[i][1]))
            if dqt[0] in encoded_pos[0] and dqt[1] in encoded_pos[1] and dqt[2] in encoded_pos[2] and dqt[3] in encoded_pos[3] and dqt[4] in encoded_pos[4]:
                counter[(dq[0],dq[1],dq[2],dq[3],dq[4])]+=1
    else:
        dq = collections.deque(maxlen=5)
        dq.append(int(arr[0]))
        dq.append(int(arr[1]))
        dq.append(int(arr[2]))
        dq.append(int(arr[3]))
        counter = collections.Counter()
        for i in range(4,arr.shape[0]):
            dq.append(int(arr[i]))
            counter[(dq[0],dq[1],dq[2],dq[3],dq[4])]+=1
    return counter        

    

cdef dict count_ngrams(str npy_path,object encoded_pos,int size,str tmpdir):
    """Searches the node in every text file and gets contexts."""
    cdef:
        list files 
        str filename 
        dict counts 
        list ng_range
        np.ndarray arr
        object counter 
        dict r 
        object k 
        int v 
    files = os.listdir(npy_path)
    counts = dict()
    for filename in files:
        # maker arr concatenating with arrays of zeros for preventing index out of bounds
        if encoded_pos == None:
            arr = np.load(npy_path + filename)[:,0:1]
        else:
            arr = np.load(npy_path + filename)[:,0:2]
        if size == 1:
            counter = unigrams(arr,encoded_pos)
        if size == 2:
            counter = bigrams(arr,encoded_pos)
        if size == 3:
            counter = trigrams(arr,encoded_pos)
        if size == 4:
            counter = quadrigrams(arr,encoded_pos)
        if size == 5:
            counter = quinquegrams(arr,encoded_pos)
        # count 
        r = dict()
        for k,v in counter.items():
            r[k]=1
            if k in counts:
                counts[k]+=v
            else:
                counts[k]=v
        with open(tmpdir + '/' + filename,'wb') as fh:
            pickle.dump(r,fh)
    return counts

cdef translate(dict counts,dict dict_words,bint lowercase,int total_files,int size,str tmpdir):
    """Translate numbers back to words in a list format for kitconc."""
    cdef:
        list ngrams
        object punct 
        dict r 
        dict d 
        dict rd 
        object a 
        object k 
        dict c 
        int v 
        int i 
        float p
        list files 
        str filename  
    ngrams = []
    files = os.listdir(tmpdir)
    # unigrams 
    if size == 1:
        punct = re.compile('^\W+$')
        # range 
        r = dict()
        for filename in files:
            with open(tmpdir + '/' + filename,'rb') as fh:
                rd = pickle.load(fh)
            d = dict()
            for a in rd:
                if lowercase == True:
                    k = dict_words[a].lower()
                else:
                    k = dict_words[a]
                if len(punct.findall(k)) == 0:
                    if k not in d:
                        d[k]=1
            for k in d:
                if k in r:
                    r[k]+=1
                else:
                    r[k]=1
        
        # ngrams
        c = dict()
        for a,v in counts.items():
            if lowercase == True:
                k = dict_words[a].lower()
            else:
                k = dict_words[a]
            if len(punct.findall(k)) == 0:
                if k in c:
                    c[k]+=v 
                else:
                    c[k]=v
        counts = None
        dict_words = None
    # multigrams
    else:
        punct = re.compile('^\W+ | \W+$| \W+ ')
        # range 
        r = dict()
        for filename in files:
            with open(tmpdir + '/' + filename,'rb') as fh:
                rd = pickle.load(fh)
            d = dict()
            for a in rd:
                if lowercase == True:
                    k = ' '.join([dict_words[i] for i in a]).lower()
                else:
                    k = ' '.join([dict_words[i] for i in a])
                if k not in d:
                    d[k]=1
            for k in d:
                if k in r:
                    r[k]+=1
                else:
                    r[k]=1
        
        # ngrams
        c = dict()
        for a,v in counts.items():
            if lowercase == True:
                k = ' '.join([dict_words[i] for i in a]).lower()
            else:
                k = ' '.join([dict_words[i] for i in a])
            if k in c:
                c[k]+=v 
            else:
                c[k]=v
        counts = None
        dict_words = None
    # make
    i = 0
    for k,v in c.items():
        if len(punct.findall(k)) == 0:
            i+=1
            p = (r[k]/float(total_files))*100
            ngrams.append((i,k,v,r[k],p))
    return ngrams 
        
def make_ngrams(str workspace,str corpus_name,object pos,int size,bint lowercase):
    cdef:
        object encoded_pos 
        int total_files 
        dict counts
        list ng_range
        list ngrams
        str tmpdir   
    tmpdir = tempfile.mkdtemp()
    encoded_pos = parse_search(workspace,corpus_name,pos)
    total_files = len(os.listdir(workspace+corpus_name + '/data/npy/'))
    counts = count_ngrams(workspace+corpus_name + '/data/npy/',encoded_pos,size,tmpdir)
    dict_words = load_dict(workspace+corpus_name+'/data/idx/words.pickle')
    ngrams = translate(counts,dict_words,lowercase,total_files,size,tmpdir)
    shutil.rmtree(tmpdir)
    return ngrams 



# Author: Hayk Aleksanyan
# create and test Trees for spatial partition



def rectArea(r):
    # returns the area of the rectangle given in min-max coordinates (top-left <--> bottom-right)
    return abs( (r[0] - r[2])*(r[1] - r[3]) )


class Node:
    """
      used to model a rectangle in quad-tree partition of the 2d plane;
      each Node models a rectangle in the partition and can have at most 4 sub-rectanlges
      representing the members of quad-tree construction
      the value of the node is a 4-tuple of integers representing a rectangle
      in "left upper-coord, and right-buttom" form
    """

    def __init__(self, val, p):
        self.value = val
        self.parent = p  # is a Node  or None if this is the root

        self.isFull = False # shows if the node has maximum number of children
                            # i.e. has reached its full capacity (2 or 4 in our case)

        self.child1 = None
        self.child2 = None
        self.child3 = None
        self.child4 = None

    def isLeaf(self):
        # true, if this node is a leaf, i.e. has no children
        return ( (self.child1 is None) and (self.child2 is None) and (self.child3 is None) and (self.child4 is None) )

    def Children(self):
        # return the list of children, if any
        c = []
        if self.child1 is not None:
            c.append(self.child1)
        if self.child2 is not None:
            c.append(self.child2)
        if self.child3 is not None:
            c.append(self.child3)
        if self.child4 is not None:
            c.append(self.child4)

        return c

    def hasLeaf_ChildrenOnly(self):
        # True, if the all children of this node (if any) are leaves
        if (self.child1 is not None):
            if self.child1.isLeaf() == False:
                return False

        if (self.child2 is not None):
            if self.child2.isLeaf() == False:
                return False

        if (self.child3 is not None):
            if self.child3.isLeaf() == False:
                return False

        if (self.child4 is not None):
            if self.child4.isLeaf() == False:
                return False

        return True


class Tree:
    """
      stores the entire quad-tree partition, where each member of the partition is a Node class
    """

    def __init__(self, root):
        # root is a Node; serves as the root of this tree
        self.root = root

    def getLeafs(self):
        #returns the leafs of the tree as a list

        if self.root == None:
            return []

        res = []

        c = self.root.Children()
        while c:
            c1 = []
            for x in c:
                if x.isLeaf():
                    res.append(x)
                else:
                    for u in x.Children():
                        c1.append(u)

            c = c1

        return res

    def nodeCount(self):
        # returns the leafs of the tree as a list

        if self.root == None:
            return 0

        res = 1
        c = self.root.Children()

        while c:
            c1 = []
            res += len(c)

            for x in c:
                #print(' '*i + 'Level ' + str(i) + ' : ' + str(x.value) )
                if x.isLeaf() == False:
                    for u in x.Children():
                        c1.append(u)

            c = c1

        return res


    def getValues(self, output = False):
        """
          traverses the tree T from the root to its leaves and returns a list of all values of all nodes
          if output == True, prints the values
        """

        if self.root == None:
            if output == True:
                print('The tree is empty')
            return []

        res = [ self.root.value ]

        c = self.root.Children()
        i = 0

        while c:
            i += 1
            c1 = []
            for x in c:
                if output == True:
                    print(' '*i + 'Level ' + str(i) + ' : ' + str(x.value) )

                res.append(x.value)
                if not x.isLeaf():
                    for u in x.Children():
                        c1.append(u)

            c = c1

        return res


    def compress(self):
        """
         compresses the tree T by removing all leaves whose siblings
         are leafs and whose parents have reached their full (child) capacity,
         i.e. have MAX number of children (2 or 4 in our case)
         performs this process from bottom-up until there is nothing to remove
        """

        if self.root == None:
            return

        current_level = [ self.root ]
        all_nodes = [ [current_level] ]

        while True:
            c = []
            for i in range( len(current_level) ):
                x = current_level[i]
                if x.child1 != None:
                    c.append(x.child1)
                if x.child2 != None:
                    c.append(x.child2)
                if x.child3 != None:
                    c.append(x.child3)
                if x.child4 != None:
                    c.append(x.child4)

            if c == []:
                break

            all_nodes.append(c)
            current_level = c[:]


        for i in range( len( all_nodes ) -1, 0, -1 ):
            for j in range( len( all_nodes[i] ) ):
                n = all_nodes[i][j]

                if n != None:
                    p = n.parent
                    if p.isFull:
                        # p has all 4 or all 2 children
                        if p.hasLeaf_ChildrenOnly():
                            # compression means that we destroy all children of a node
                            # with only leaf children and where all child nodes are occupied
                            p.child1 = None
                            p.child2 = None
                            p.child3 = None
                            p.child4 = None



    def areaCovered( self ):
        """
          compute the numerical value of the 2d area covered by this Tree
          the object represented by this tree is the disjoin union of its leaves;
          leaves are rectangles, thus we need to compute the sum of the areas of these rectangles
        """

        a = 0
        c = self.getLeafs()

        for r in c:
            a += rectArea(r.value)

        return a



# Author: Hayk Aleksanyan
# create bounding hierarchical boxes for word shapes

from kitconc.wordle_trees import Node, Tree



def intersectionRect(r1, r2, shift1 = (0,0), shift2 = (0,0), extraSize = 3 ):
    """
     gets two 4-tuples of integers representing a rectangle in min,max coord-s
     optional params. @shifts can be used to move boxes on a larger canvas (2d plane)
                      @extraSize, forces the rectangles to stay away from each other by the given size (number of pixels)

     returns True if the rectangles intersect
    """

    if ((min(r1[0] - extraSize + shift1[0] , r1[2] + extraSize + shift1[0] ) > max( r2[0] - extraSize + shift2[0], r2[2] + extraSize + shift2[0] ) )
        or ( max(r1[0] - extraSize + shift1[0], r1[2] + extraSize + shift1[0] ) < min( r2[0] - extraSize + shift2[0], r2[2] + extraSize + shift2[0] ) ) ):
        return False

    if ((min(r1[1] - extraSize + shift1[1] , r1[3] + extraSize + shift1[1] ) > max( r2[1] - extraSize + shift2[1], r2[3] + extraSize + shift2[1]) )
        or ( max(r1[1] - extraSize + shift1[1], r1[3]  + extraSize + shift1[1] ) < min( r2[1] - extraSize + shift2[1], r2[3]+ extraSize + shift2[1] ) ) ):
        return False

    return True


def getQuadTree(im, minW, minH):
    """
      returns the quad-tree representation of the image @im, i.e.
      a Tree, where the value of each node is a 4-tuple of ints (can be 2 for some of the leafs),
      representing min-max of hierarchic boxes
      here minW and minH are the width, height of the minimal box
    """

    box_0 = im.getbbox()  # the initial box

    p = Node(box_0, None) # the root of the tree
    T = Tree( p )

    stack = [ p ]  # (4 ints, a tuple)

    W, H = abs(box_0[0] - box_0[2]), abs(box_0[1] - box_0[3])

    if ( (H <= minH) and (W <= minW) ):
        # we do not split further if the height and width are small enough
        return T


    while stack:
        x = stack.pop()
        x_box = x.value    # the box coordinates
        full_node = True   # shows if the x is full

        W, H = abs(x_box[0] - x_box[2]), abs(x_box[1] - x_box[3])
        if ( (H <= minH) and (W <= minW)):
            continue

        #consider the 4 sub-boxes:

        if (x_box[0] + x_box[2])%2 == 0:
            d1 = (x_box[0] + x_box[2])>>1
        else:
            d1 = (x_box[0] + x_box[2] + 1)>>1

        if (x_box[1] + x_box[3])%2 == 0:
            d2 = (x_box[1] + x_box[3])>>1
        else:
            d2 = (x_box[1] + x_box[3] + 1)>>1

        # (x0, x1, d1, x3), (d1+1, x1, x2, d2), (x0, d2+1, d1, x3), (d1+1, d2+1, x2, x3)

        #we now set the children of x
        if ((H > minH) and (W > minW) ):
            #we need 4 sub-rectangles

            x_child = (x_box[0], x_box[1], d1, d2 )
            if im.crop(x_child).getbbox() is not None:
                x.child1 = Node( x_child, x )
                stack.append(x.child1)
            else:
                full_node = False

            x_child = (d1 , x_box[1], x_box[2], d2)
            if im.crop(x_child).getbbox() is not None:
                x.child2 = Node( x_child,  x )
                stack.append(x.child2)
            else:
                full_node = False

            x_child = (x_box[0], d2, d1, x_box[3])
            if im.crop(x_child).getbbox() is not None:
                x.child3 = Node( x_child,  x )
                stack.append(x.child3)
            else:
                full_node = False

            x_child = (d1 , d2 , x_box[2], x_box[3])
            if im.crop(x_child).getbbox() is not None:
                x.child4 = Node( x_child,  x )
                stack.append(x.child4)
            else:
                full_node = False

            x.isFull = full_node

        else:
            if (( H <= minH  ) and ( W > minW ) ): # don't split the y-coord, but only x

                x_child = (x_box[0], x_box[1], d1, x_box[3] )
                if im.crop(x_child).getbbox() is not None:
                    x.child1 = Node( x_child, x )
                    stack.append(x.child1)
                else:
                    full_node = False


                x_child = (d1, x_box[1], x_box[2], x_box[3] )
                if im.crop(x_child).getbbox() is not None:
                    x.child2 = Node( x_child, x )
                    stack.append(x.child2)
                else:
                    full_node = False

                x.isFull = full_node

            else: #we're in a position that we only split H

                x_child = (x_box[0], x_box[1], x_box[2], d2 )
                if im.crop(x_child).getbbox() is not None:
                    x.child1 = Node( x_child, x )
                    stack.append(x.child1)
                else:
                    full_node = False

                x_child = (x_box[0], d2, x_box[2], x_box[3] )
                if im.crop(x_child).getbbox() is not None:
                    x.child2 = Node( x_child, x )
                    stack.append(x.child2)
                else:
                    full_node = False


                x.isFull = full_node

    return T


def insideCanvas( T, shift, canvas_size ):
    """
     @T is the tree-representation of a word,
     @shift is the word's bounding box's upper-left corner coord on canvas
     @canvas_size is a tuple (width, height) of the canvas

     returns True if the word's leaves stay inside the canvas
    """

    stack = [ T.root ]
    W, H = canvas_size
    sh_w, sh_h = shift


    while stack:
        v = stack.pop()

        a, b, c, d = v.value # a 4 tuple, left upper-coord and right-buttom coordinates
        if ( (a + sh_w < 0) or ( c + sh_w > W ) or (b + sh_h < 0) or (d + sh_h > H) ) == False:
            continue

        if v.isLeaf():
            return False

        stack += v.Children()

    return True



def collisionTest(T1, T2, shift1, shift2, stay_away):
    """
       the input is a pair of trees representing the objects as a quad-tree
       shift_1 = (a1, b1) and shift2 = (a2, b2) are the left-top coordinats of the boxes on the large canvas
       this means that all boxes in T_i must be shifted by (a_i, b_i), where i = 1, 2

       @stay_away parameter forces bounding boxes to stay at least @stay_away pixels away from each other

       return True iff the quad-trees have intersecting leaves, meaning the images they respresent actually intersect
    """

    r1, r2 = T1.root, T2.root

    if ( (not r1) or (not r2) ):
        return False


    stack = [ (r1, r2) ]

    while stack:
        p1, p2 = stack.pop()   # the nodes of the 1st and the 2nd trees

        if intersectionRect(p1.value, p2.value, shift1, shift2, stay_away) == False:
            # if larger rectangles do not collide, their children will not collide either
            # hence no need to go for sub-nodes
            continue

        if ( p1.isLeaf() and p2.isLeaf() ):
            # leaves collide, we're done
            return True


        c1, c2 = p1.Children(), p2.Children()

        if not c1:
            for x in c2:
                stack.append( (p1, x) )
        else:
            if not c2:
                for x in c1:
                    stack.append( (x, p2) )
            else:
                # none are empty, i.e. the nodes are not leaves
                for x in c1:
                    for y in c2:
                        stack.append( (x, y) )


    return False



import os 
import numpy as np 
import pickle 
import re 

def load_dict(dict_path):
    """Loads a dictionary in pickle format."""
    with open(dict_path,'rb') as fh:
        d = pickle.load(fh)
    return d 

def parse_search(workspace,corpus_name,node,pos,lowercase):
    """Parses the input search and returns a tuple with encoded word and pos node.
    """
    # load dictionaries for encoding words to numbers
    if lowercase == False:
        dict_words = load_dict(workspace+corpus_name+'/data/idx/wordslc.pickle') # lowercase
    else:
        dict_words = load_dict(workspace+corpus_name+'/data/idx/words.pickle')
        dict_words = {v: k for k, v in dict_words.items()}
    dict_tags = load_dict(workspace+corpus_name+'/data/idx/tags.pickle')
    dict_tags = {v: k for k, v in dict_tags.items()}
    # parse node
    n = []
    if lowercase == False:
        node = node.lower()
    for w in node.strip().split(' '):
        if len(w.strip()) != 0:
            n.append(w.strip())
            break # allow 1 word only
    # parse pos 
    if pos != None:
        p = []
        if type(pos) == list:
            for t in pos:
                if len(t.strip())!= 0:
                    p.append(t.strip())
                    break # allow 1 tag only
        else:
            for t in pos.strip().split(' '):
                if len(t.strip())!=0:
                    p.append(t.strip())
                    break # allow 1 tag only
    # encode node
    encoded_node = []
    for w in n:
        if w in dict_words:
            arr = np.array(dict_words[w],dtype=int)
            encoded_node.append(arr)
        else: # if any word has no match, return None
            encoded_node = None
            break
    
    # encode pos
    if pos != None:
        encoded_pos = []
        for t in p:
            if t in dict_tags:
                arr = np.array(dict_tags[t],dtype=int)
                encoded_pos.append(arr)
            else: # if any tag has no match, return None
                encoded_pos = None
                break 
    else:
        encoded_pos = None
    return (encoded_node, encoded_pos)


def search_node(npy_path,encoded_node,encoded_pos,size):
    """Searches the node in every text file and gets contexts."""
    files = os.listdir(npy_path)
    contexts = []
    for filename in files:
        # maker arr concatenating with arrays of zeros for preventing index out of bounds
        arr = np.concatenate([np.zeros((size+size,4),dtype=int),np.load(npy_path + filename),np.zeros((size+size,4),dtype=int)])
        # get matching indexes
        ix = arr_search_indexes(arr,encoded_node,encoded_pos)
        # get contexts from indexes
        contexts += get_context(arr,ix,size)
    return contexts 


def arr_search_indexes(arr,search_w,search_t):
    """Searches the node in arr and returns the matching indexes. (Max. 3 words)"""
    size = len(search_w)
    if search_t == None:
        # search only words
        if size == 1:
            ix = np.isin(arr[:,[0]],search_w)
            ix = np.where(ix)[0]
    else:
        # search words and POS
        if size == 1:
            iw = np.isin(arr[:,[0]],search_w)
            it = np.isin(arr[:,[1]],search_t)
            ix = np.where(iw & it)[0]
    return ix 

def get_context(arr,indexes,size):
    """Gets the contexts from indexes."""
    d = dict()
    if size == 1:
        for i in indexes:
            l = (arr[i,0])
            if l in d:
                d[l]+=1
            else:
                d[l]=1
    elif size == 2:
        for i in indexes:
            l = (arr[i-1,0],arr[i,0])
            if l in d:
                d[l]+=1
            else:
                d[l]=1
            l = (arr[i,0],arr[i+1,0])
            if l in d:
                d[l]+=1
            else:
                d[l]=1
    elif size == 3:
        for i in indexes:
            l = (arr[i-2,0],arr[i-1,0],arr[i,0])
            if l in d:
                d[l]+=1
            else:
                d[l]=1
            l = (arr[i-1,0],arr[i,0],arr[i+1,0])
            if l in d:
                d[l]+=1
            else:
                d[l]=1
            l = (arr[i,0],arr[i+1,0],arr[i+2,0])
            if l in d:
                d[l]+=1
            else:
                d[l]=1
    return [d] 
        

def translate(contexts,dict_words,lowercase,total_files,size):
    """Translate numbers back to words in a list format for kitconc KIWC object."""
    clusters = []
    r = dict()
    c = dict() 
    n = dict()
    punct = re.compile('^\W+ | \W+$| \W+ ')
    for d in contexts:
        n=dict()
        for k,v in d.items():
            # translate clusters
            if size == 1:
                if k !=0:
                    if lowercase == True:
                        w = dict_words[k].lower()
                    else:
                        w = dict_words[k]
            else:
                if 0 not in k:
                    if lowercase == True:
                        w = ' '.join([dict_words[i].lower() for i in k])
                    else:
                        w = ' '.join([dict_words[i] for i in k])
            # add translated to new dict
            if len(punct.findall(w)) == 0:
                if w in n:
                    n[w]+=v
                else:
                    n[w]=v
        # count range and append counts
        for w,v in n.items():
            if w in r:
                r[w]+=1
            else:
                r[w]=1
            if w in c:
                c[w]+=v 
            else:
                c[w]=v
    
    contexts = None
    # make clusters
    i = 0 
    for w,v in c.items():
        i+=1
        p = (r[w]/float(total_files))*100
        clusters.append((i,w,v,r[w],p))
    return clusters 
    
def make_clusters(workspace,corpus_name,node,pos,lowercase,size):
    encoded_node,encoded_pos = parse_search(workspace,corpus_name,node,pos,lowercase)
    if encoded_node != None: # search node is OK
        contexts = search_node(workspace+corpus_name + '/data/npy/',encoded_node,encoded_pos,size)
        if len(contexts)!=0:
            total_files = len(os.listdir(workspace+corpus_name + '/data/npy/'))
            dict_words = load_dict(workspace+corpus_name+'/data/idx/words.pickle')
            clusters = translate(contexts,dict_words,lowercase,total_files,size)
        else:
            clusters=[]
    else: # no match in search node, set empty
        clusters = []
    return clusters 
    
    
    
    


import os 
import numpy as np 
import pickle 
import re 

def load_dict(dict_path):
    """Loads a dictionary in pickle format."""
    with open(dict_path,'rb') as fh:
        d = pickle.load(fh)
    return d 

def parse_search(workspace,corpus_name,node,pos,case_sensitive,regexp):
    """Parses the input search and returns a tuple with encoded word and pos node.
    """
    # load dictionaries for encoding words to numbers
    if case_sensitive == False:
        dict_words = load_dict(workspace+corpus_name+'/data/idx/wordslc.pickle') # lowercase
    else:
        dict_words = load_dict(workspace+corpus_name+'/data/idx/words.pickle')
        dict_words = {v: k for k, v in dict_words.items()}
    dict_tags = load_dict(workspace+corpus_name+'/data/idx/tags.pickle')
    dict_tags = {v: k for k, v in dict_tags.items()}
    # parse node
    n = []
    if case_sensitive == False:
        node = node.lower()
    for w in node.strip().split(' '):
        if len(w.strip()) != 0:
            n.append(w.strip())
    # parse pos 
    if pos != None:
        p = []
        if type(pos) == list:
            for t in pos:
                if len(t.strip())!= 0:
                    p.append(t.strip())
        else:
            for t in pos.strip().split(' '):
                if len(t.strip())!=0:
                    p.append(t.strip())
    # encode node
    encoded_node = []
    if regexp == False:
        for w in n:
            if w in dict_words:
                arr = np.array(dict_words[w],dtype=int)
                encoded_node.append(arr)
            else: # if any word has no match, return None
                encoded_node = None
                break
    else: # use regexp:
        patterns = []
        for ptrn in n:
            patterns.append(re.compile(ptrn))
        for ptrn in patterns:
            a = []
            for k in dict_words:
                if ptrn.match(k) != None:
                    a+=dict_words[k]
            if len(a)!=0:
                arr = np.array(a,dtype=int)
                encoded_node.append(arr)
            else: # if any pattern has no match, return None
                encoded_node = None
                break 
    # encode pos
    if pos != None:
        encoded_pos = []
        for t in p:
            if t in dict_tags:
                arr = np.array(dict_tags[t],dtype=int)
                encoded_pos.append(arr)
            else: # if any tag has no match, return None
                encoded_pos = None
                break 
    else:
        encoded_pos = None
    return (encoded_node, encoded_pos)


def search_node(npy_path,encoded_node,encoded_pos):
    """Searches the node in every text file and gets positions."""
    files = os.listdir(npy_path)
    data = []
    for filename in files:
        # load arr 
        arr = np.load(npy_path + filename)
        # get matching indexes
        ix = arr_search_indexes(arr,encoded_node,encoded_pos)
        positions = []
        for i in ix.tolist():
            p = round((i / arr.shape[0])*100,2)
            positions.append(p)
        # get positions
        data.append((filename.replace('.npy',''),arr.shape[0],len(ix),positions))
    return data


def arr_search_indexes(arr,search_w,search_t):
    """Searches the node in arr and returns the matching indexes. (Max. 3 words)"""
    size = len(search_w)
    if search_t == None:
        # search only words
        if size == 1:
            ix = np.isin(arr[:,[0]],search_w)
            ix = np.where(ix)[0]
        elif size == 2:
            # w1
            x1 = np.isin(arr[:,[0]],search_w[0])
            if np.any(x1)==False:
                return np.where(x1)[0]
            # w2 
            x2 = np.isin(arr[:,[0]],search_w[1])
            if np.any(x2)==False:
                return np.where(x2)[0]
            # match indexes
            x = np.arange(0,x1.shape[0]-1)
            y = np.arange(1,x1.shape[0])
            # return indexes
            ix = np.where(x1[x] & x2[y])[0]
            return ix 
        elif size == 3:
            # w1
            x1 = np.isin(arr[:,[0]],search_w[0])
            if np.any(x1)==False:
                return np.where(x1)[0]
            # w2
            x2 = np.isin(arr[:,[0]],search_w[1])
            if np.any(x2)==False:
                return np.where(x2)[0]
            # w3
            x3 = np.isin(arr[:,[0]],search_w[2])
            if np.any(x3)==False:
                return np.where(x3)[0]
            # match indexes
            x = np.arange(0,x1.shape[0]-2)
            y = np.arange(1,x1.shape[0]-1)
            z = np.arange(2,x1.shape[0])
            # return indexes
            ix = np.where(x1[x] & x2[y] & x3[z])[0]
    else:
        # search words and POS
        if size == 1:
            iw = np.isin(arr[:,[0]],search_w)
            it = np.isin(arr[:,[1]],search_t)
            ix = np.where(iw & it)[0]
        elif size == 2:
            # w1 t1
            x1 = np.isin(arr[:,[0]],search_w[0])
            t1 = np.isin(arr[:,[1]],search_t[0])
            if np.any(x1)==False:
                return np.where(x1)[0]
            if np.any(t1)==False:
                return np.where(t1)[0]
            # w2 t2 
            x2 = np.isin(arr[:,[0]],search_w[1])
            t2 = np.isin(arr[:,[1]],search_t[1])
            if np.any(x2)==False:
                return np.where(x2)[0]
            if np.any(t2)==False:
                return np.where(t2)[0]
            # match indexes
            x = np.arange(0,x1.shape[0]-1)
            y = np.arange(1,x1.shape[0])
            # return indexes 
            ix = np.where(x1[x] & t1[x] & x2[y] & t2[y])[0]
        elif size == 3:
            # w1 t1
            x1 = np.isin(arr[:,[0]],search_w[0])
            t1 = np.isin(arr[:,[1]],search_t[0])
            if np.any(x1)==False:
                return np.where(x1)[0]
            if np.any(t1)==False:
                return np.where(t1)[0]
            # w2 t2
            x2 = np.isin(arr[:,[0]],search_w[1])
            t2 = np.isin(arr[:,[1]],search_t[1])
            if np.any(x2)==False:
                return np.where(x2)[0]
            if np.any(t2)==False:
                return np.where(t2)[0]
            # w3 t3
            x3 = np.isin(arr[:,[0]],search_w[2])
            t3 = np.isin(arr[:,[1]],search_t[2])
            if np.any(x3)==False:
                return np.where(x3)[0]
            if np.any(t3)==False:
                return np.where(t3)[0]
            # match indexes
            x = np.arange(0,x1.shape[0]-2)
            y = np.arange(1,x1.shape[0]-1)
            z = np.arange(2,x1.shape[0])
            # return indexes
            ix = np.where(x1[x] & t1[x] & x2[y] & t2[y] & x3[z] & t3[z])[0]
    return ix 


def translate(positions):
    """Translate numbers back to words in a list format for kitconc KIWC object."""
    dispersion = []
    dpts = dict()
    i = 0
    total_s1 = 0
    total_s2 = 0
    total_s3 = 0
    total_s4 = 0
    total_s5 = 0
    
    for d in positions:
        i+=1
        s1 = 0
        s2 = 0 
        s3 = 0
        s4 = 0
        s5 = 0
        dpts[d[0]] = d[3]
        for point in d[3]:
            p = round(point)
            if p >= 0 and p <= 19:
                s1+=1
            elif p >= 20 and p <= 39:
                s2+=1
            elif p >= 40 and p <= 59:
                s3+=1
            elif p >= 60 and p <= 79:
                s4+=1
            elif p >= 80 and p <= 100:
                s5+=1
        total_s1 += s1
        total_s2 += s2
        total_s3 += s3
        total_s4 += s4
        total_s5 += s5
        dispersion.append((i ,d[0],d[1],d[2],s1,s2,s3,s4,s5))
    return ((total_s1,total_s2,total_s3,total_s4,total_s5),dpts,dispersion)
        
    
def make_dispersion(workspace,corpus_name,node,pos,case_sensitive,regexp):
    dispersion=[]
    dpts = dict()
    totals = tuple()
    encoded_node,encoded_pos = parse_search(workspace,corpus_name,node,pos,case_sensitive,regexp)
    if encoded_node != None: # search node is OK
        positions = search_node(workspace+corpus_name + '/data/npy/',encoded_node,encoded_pos)
        if len(positions)!=0:
            totals,dpts,dispersion = translate(positions)
    return (totals,dpts,dispersion)
    
    
    
    